/*
signed_dist.c
This program computes an optimal sorting by reversal for signed permutations.
The code is not documented and not very user-friendly. We apologize for that. Moreover this is not the
fastest (definitely the easiest)implementation of the polynomial algorithm
recently proposed by Hannenhalli and Pevzner.
This is a Plan 9 APE port by Peter A. Cejchan  <cej@gli.cas.cz | pac7@post.cz> 
Thu Jul 17 16:34:32 EDT 2003
Supported from grant #GA205/03/1124.
Changes in sources (if any) marked  '++pac'. Type 'mk install' to compile and install. Enjoy!


usage: signed_dist size source_permutation target_permutation
(whitespace delimited signed numbers)
It takes two signed permuations from stdin and outputs their reversal distance
and an optimal scenario (sequence of permutations).


*/


#include<stdio.h> 
#include<assert.h> 
#define MAX_SIZE 500
#define COMPONENT 1
#define ORIENTED 1
#define UNORIENTED 2
#define HURDLE 3
#define INFINITY 10000
#define NEGINFINITY -10000

#define MIN(x,y) ((x)>(y)?(y):(x))
#define MAX(x,y) ((x)<(y)?(y):(x))

int g_perm[MAX_SIZE], store_perm[MAX_SIZE];
int source[MAX_SIZE], target[MAX_SIZE];
int is_interval[MAX_SIZE+2][MAX_SIZE+2];
int SIZE;
int *reached[2];

/*---------------------------*/
void print_perm(int *perm, int size)
{
int i;
for(i=0;i<size;i++)
	printf("%d ",perm[i]);
printf("\n");
//++pac 
fflush(stdout);
}

/*---------------------------*/
void copy(int *source, int *target, int size)
{
int i;
for(i=0;i<size;i++)
	target[i] = source[i];
}

/*-----------------------------*/
/*
*	take an index and a size. compute p = inv_index(index,size).
*   p' = p appended by a 0;
*   compute q' by reversing p' from i to j.
*   q = q' after rotating to get 0 at the end.
*   return get_index(q,size)
*/
void reverse(int *perm, int size, int start, int end)
{
int i, zero_pos = -1;
int dir = 1;
assert(end >= start);
while((end-start) > 0)
{
int temp = perm[start];
	perm[start] = -1*perm[end];
	perm[end] = -1*temp;
	if(dir == 1)
	start++;
	end--;
}
if(start == end) /* odd size */
	perm[start] *= -1;
}

/*--------------- compute #breakpoints---------------*/
int num_breakpoints(int *perm, int size)
{
int _b = 0;
int i;
if(perm[0] != 1) _b++;
if(perm[size-1] != size) _b++;
for(i=0;i<size-1;i++)
	if(perm[i+1]-perm[i] != 1) _b++;
return _b;
}
/*--------------- compute #cycles-------------------*/
int num_cycles(int *perm, int size)
{
int u_perm[2*MAX_SIZE+2];
int done[2*MAX_SIZE+2];
int black_edge[2*MAX_SIZE+2];
int grey_edge[2*MAX_SIZE+2];
int i, j;
int _c = 0;

		/* convert to unsigned */

u_perm[0] = 0;
u_perm[2*size+1] = 2*size+1;
for(i=0;i<size;i++)
{
	if(perm[i] > 0)
	{
		u_perm[2*i+1] = 2*perm[i]-1;
		u_perm[2*i+2] = 2*perm[i];
	}
	else
	{
		u_perm[2*i+1] = -2*perm[i];
		u_perm[2*i+2] = -2*perm[i]-1;
	}
}

for(i=0;i<2*size+2;i++)
{
	done[i] = 0;
	black_edge[i] = -1;
	grey_edge[i] = -1;
}
	/* set black -edges */
for(i=0;i<2*size+1;i++)
	if(abs(u_perm[i+1]-u_perm[i]) != 1)
	{
		black_edge[i] = i+1;
		black_edge[i+1] = i;
	}

	/* set grey -edges */
for(i=0;i<2*size;i++)
	for(j=i+2;j<2*size+2;j++)
		if(abs(u_perm[j]-u_perm[i]) == 1)
		{
			grey_edge[i] = j;
			grey_edge[j] = i;
		}






for(i=0;i<2*size+2;i++)
{
	if(done[i] == 0 && black_edge[i] != -1 && grey_edge[i] != -1)
	{
	int next;
		done[i] = 1;
		next = i;
		do
		{
			next = black_edge[next];
			done[next] = 1;
			next = grey_edge[next] ;
			done[next] = 1;
		}
		while(next != i);
		_c++;

	}
}

return _c;
}

/*--------------- identify bad configuration-------------------*/
/* compute number of hurdles also and return 1000*hurdle+fortress */
/*	Return 1 if the permutation has a bad configuration	*/
int is_bad_config(int *perm, int SIZE)
{
int i,j,k, _z = 0, size = SIZE, fortress = 1; /* fortress is bad_config, 1 by default */
int entire_unoriented = 0, outermost_is_hurdle = 0;
int t_perm[MAX_SIZE+2];
int back_perm[MAX_SIZE+2];
int leftmost_hurdle = INFINITY, rightmost_hurdle = NEGINFINITY;
int left_inner=-1, right_inner=size+2;
t_perm[0] = 1;
t_perm[size+1] = size+2;
for(i=0;i<size;i++)
{
	if(perm[i] > 0)
		t_perm[i+1] = perm[i]+1;
	else
		t_perm[i+1] = perm[i]-1;
}

size += 2;
copy(t_perm, back_perm, size);
for(i=0;i<size;i++)
	for(j=i+2;j<size;j++)
		is_interval[i][j] = 0;
	/* mark intervals */
for(i=0;i<size;i++)
	for(j=i+2;j<size;j++)
	{
	int min, max, k, good = 1;
		if(	(t_perm[i+1]-t_perm[i] == 1 || t_perm[j]-t_perm[j-1] == 1) ||
		  	(t_perm[i] * t_perm[j] < 0) ||
			(t_perm[i] > t_perm[j]) )
			continue;

		min = MIN(abs(t_perm[i]), abs(t_perm[j]));
		max = MAX(abs(t_perm[i]), abs(t_perm[j]));
		for(k=i+1; k<j; k++)
			if(abs(t_perm[k]) > max || abs(t_perm[k]) < min)
			{
				good = 0;
				break;
			}
		if(good && max-min == j-i)
		{
			is_interval[i][j] = COMPONENT;
			break;
		}
	}

	/* scan the intervals and identify the bad ones */
	/* also keep track if the outermost component is bad */
	for(k=2;k<size;k++)
	for(i=0;i<size-k;i++)
	{
	int m, n;
		if(is_interval[i][i+k] == 0)
			continue;
		/* check if (i, i+k) is UNORIENTED */
		is_interval[i][i+k] = UNORIENTED; /* by default */
		for(j=i;j<i+k;j++)
			if(t_perm[j]*t_perm[j+1] < 1)
			{
				is_interval[i][i+k] = ORIENTED;
				break;
			}
			/* sort the COMPONENT so that the containing components go through the
				above test properly */	
		for(j=i+1;j<i+k;j++)
			t_perm[j] = t_perm[i] + j - i;

		if(is_interval[i][i+k] == ORIENTED)
			continue;

		/* check if it is HURDLE */
		is_interval[i][i+k] = HURDLE; /* by default */
		for(m=i+1;m<i+k-2 && is_interval[i][i+k] == HURDLE;m++)
		for(n=m+2;n<i+k && is_interval[i][i+k] == HURDLE;n++)
			if(is_interval[m][n] == HURDLE)
				is_interval[i][i+k] = UNORIENTED; 
		if(is_interval[i][i+k] == HURDLE) /* THIS SHOULD HAVE BEEN UNORIENTED */
		{
			_z++;
			leftmost_hurdle = MIN(leftmost_hurdle, i);
			rightmost_hurdle = MAX(rightmost_hurdle, i+k);
		}
	}

	if(_z == 0)
		return 0;
	/* check if all HURDLEs are contained in a UNORIENTED */
	for(i=0;i<leftmost_hurdle && entire_unoriented == 0;i++)
	for(j=size-1;j>rightmost_hurdle && entire_unoriented == 0;j--)
		if(is_interval[i][j] == UNORIENTED)
		{
		int m,n,found = 0;
			entire_unoriented = 1; /* to break the loop */
			/* check if it is in a castle */
			/* Go outwards so as to identify the innermost unoriented covering */
			/* all other hurdles */
			for(m=leftmost_hurdle-1;m>i && found==0;m--)
			for(n=rightmost_hurdle+1;n<j && found==0;n++)
				if(is_interval[m][n] == UNORIENTED)
				{
					found = 1;
					left_inner = m;
					right_inner = n;
				}
			if(found == 0) 
			{
				int is_hurdle = 1, prev_right = -1;
				fortress = 0; /* can't be a bad configuration */
				left_inner = i;
				right_inner = j;
					/* check if it separates two hurdles in which case it is not
						a HURDLE itself */
				for(m=left_inner+1;m<right_inner && prev_right == -1;m++)
				for(n=right_inner-1;n>m+1 && prev_right == -1;n--)
					if(is_interval[m][n] == UNORIENTED || is_interval[m][n] == HURDLE)
						prev_right = n;
				if(prev_right != -1) /* there is something inside */
				{
					while(prev_right<right_inner-1 && is_hurdle == 1)
					{
					int curr_left = -1, curr_right;
						for(m=prev_right;m<right_inner && curr_left == -1;m++)
						for(n=right_inner-1;n>m+1 && curr_left == -1;n--)
							if(is_interval[m][n] == UNORIENTED || is_interval[m][n] == HURDLE)
							{
								curr_left = m;
								curr_right = n;
							}
						if(curr_left != -1)
							/* check if there is a breakpoint between the two 
								which must beint to the innermost UCCC */
						{
							for(m=prev_right;m<curr_left;m++)
								if(back_perm[m+1] != back_perm[m]+1) /*break point */
								{
									is_hurdle = 0;
									break;
								}
							prev_right = curr_right;
						}
						else break;
					}
				}
				if(is_hurdle == 1)
				{
					_z++;
					is_interval[i][j] = HURDLE;
				}
			}
			else 
			{
					/* check if there is something between the innermost and the outermost
					in which case the outermost hurdle is in castle */
			int between = 0;
				_z++;
				is_interval[i][j] = HURDLE;
				for(m=i+1;m<left_inner && between == 0;m++)	
				for(n=j-1;n>right_inner && between == 0;n--)	
					if(is_interval[m][n] == UNORIENTED)
						between = 1;
				if(between == 0)
					/* we need to check if the inner one separates two hurdles */
				{
				int prev_right = -1;
					for(m=left_inner+1;m<right_inner && prev_right == -1;m++)
					for(n=right_inner-1;n>m+1 && prev_right == -1;n--)
						if(is_interval[m][n] == UNORIENTED || is_interval[m][n] == HURDLE)
							prev_right = n;
					if(prev_right != -1) /* there is something inside */
					{
						while(prev_right<right_inner-1 && fortress == 1)
						{
						int curr_left = -1, curr_right;
							for(m=prev_right;m<right_inner && curr_left == -1;m++)
							for(n=right_inner-1;n>m+1 && curr_left == -1;n--)
								if(is_interval[m][n] == UNORIENTED || is_interval[m][n] == HURDLE)
								{
									curr_left = m;
									curr_right = n;
								}
							if(curr_left != -1)
								/* check if there is a breakpoint between the two 
									which must beint to the innermost UCCC */
							{
								for(m=prev_right;m<curr_left;m++)
									if(back_perm[m+1] != back_perm[m]+1) /*break point */
									{
										fortress = 0;
										break;
									}
								prev_right = curr_right;
							}
							else break;
						}
					}
				}
			}
		}

	/* check for bad configuration */
	/* fortress could already be 0 */
	if(_z<3 || _z%2 ==  0) fortress = 0;
	/* for each hurdle check if it is in a castle */
	for(i=left_inner+1;i<right_inner && fortress==1;i++)
	for(j=i+2;j<right_inner && fortress==1;j++)
		if(is_interval[i][j] == HURDLE)
		{
		int m,n,contained = 0;
			/* check if there is a unoriented covering this one */
			/* go outwards to detect the smallest such component */
			for(m=i-1;m>left_inner && fortress==1 && contained == 0;m--)
			for(n=j+1;n<right_inner && fortress==1 && contained == 0;n++)
				if(is_interval[m][n] == UNORIENTED)
				/* check if that interval covers some other HURDLE */
				{
				int p,q;
					contained = 1;
					for(p=m+1;p<n && fortress==1;p++) 
					for(q=p+2;q<n && fortress==1;q++) 
								/* we could save time by by checking*/
								/* only to left and right of hurdle(i,j) */
						if(	is_interval[p][q] == HURDLE && p != i )
							fortress = 0;
				}
			if(	fortress == 1 && contained == 0 && 
				left_inner < INFINITY && /* redundant */ right_inner > NEGINFINITY &&
				((i == leftmost_hurdle) || (j == rightmost_hurdle) ) 
			  )

				/* check if this is the leftmost or the rightmost hurdle and all the 
					other hurdles are covered by a unoriented components
					not separating any two hurdles. */
			{
			int p,q,left_boundary, right_boundary, entire_unoriented = 0;
			int local_leftmost = INFINITY, local_rightmost = NEGINFINITY;
				if(i == leftmost_hurdle)
				{
				int r,s;
					for(r=i+2;r<size-1;r++)
					for(s=r+2;s<size+1;s++)
						if(is_interval[r][s] == HURDLE)
						{
							local_leftmost = r;
							r=s=size+1; /* to exit */
						}
					local_rightmost = rightmost_hurdle;
				}
				else if(j == rightmost_hurdle)
				{
				int r,s;
					for(s=j-2;s>1;s--)
					for(r=s-2;r>-1;r--)
						if(is_interval[r][s] == HURDLE)
						{
							local_rightmost = s;
							r=s=-1; /* to exit */
						}
					local_leftmost = leftmost_hurdle;
				}	
				/* check if the rest is contained in one unoriented component */
				/* check if all HURDLEs are contained in a UNORIENTED */
				for(p=i+2;p<local_leftmost && entire_unoriented == 0;p++)
				for(q=size-1;q>local_rightmost && entire_unoriented == 0;q--)
					if(is_interval[p][q] == UNORIENTED)
					{
					int m,n,found = 0;
						entire_unoriented = 1; /* to break the loop */
								/* we need to check if this one separates two hurdles */
						{
						int prev_right = -1;
							for(m=p+1;m<q && prev_right == -1;m++)
							for(n=q-1;n>m+1 && prev_right == -1;n--)
								if(is_interval[m][n] == UNORIENTED || is_interval[m][n] == HURDLE)
									prev_right = n;
							if(prev_right != -1) /* there is something inside */
							{
								while(prev_right<q-1 && fortress == 1)
								{
								int curr_left = -1, curr_right;
									for(m=prev_right;m<q && curr_left == -1;m++)
									for(n=q-1;n>m+1 && curr_left == -1;n--)
										if(	is_interval[m][n] == UNORIENTED || 
											is_interval[m][n] == HURDLE)
										{
											curr_left = m;
											curr_right = n;
										}
									if(curr_left != -1)
										/* check if there is a breakpoint between the two 
											which must beint to the innermost UCCC */
									{
										for(m=prev_right;m<curr_left;m++)
											if(back_perm[m+1] != back_perm[m]+1) /*break point */
											{
												fortress = 0;
												break;
											}
										prev_right = curr_right;
									}
									else break;
								}
							}
						}
					}
				if(entire_unoriented == 0) fortress = 0;
			}

		i = j; /* just to save time since we know hurdles are linear */
		}
return (1000*_z+fortress);
}
/*------------just separate  _z and fortress-----------------*/
void get_z_fortress(int *perm, int SIZE)
{
int _z_fortress = is_bad_config(perm, SIZE);
int _z = _z_fortress/1000;
int _fortress = _z_fortress - 1000*_z;
printf("_z = %d, _fortress = %d",_z,_fortress);
}

/* reversal_dist -------------------------*/
void reversal_dist(int *perm, int size)
{
int curr_b = num_breakpoints(perm, size), 
	curr_c = num_cycles(perm, size), 
	curr_z_fortress = is_bad_config(perm, size),
	curr_z = curr_z_fortress/1000,
	curr_fortress = curr_z_fortress - 1000*curr_z;
int i,j;
int  curr_perm[MAX_SIZE], temp_perm[MAX_SIZE];
int next_b, next_c, next_z, next_z_fortress, next_fortress;
int store_b = curr_b, store_c = curr_c, store_z = curr_z, 
	store_z_fortress = curr_z_fortress, store_fortress = curr_fortress;
int lower_bound = curr_b + curr_z - curr_c + curr_fortress;
int rev_dist = 0;
int found;

copy(perm, curr_perm, size);

printf("\n\nAn optimal reversal scenario:\n\n");
//++pac 
fflush(stdout);
while(curr_b > 0) 
{
	/*
	printf("(%d-%d+%d+%d = %d)", 
			curr_b,curr_c,curr_z,curr_fortress,curr_b-curr_c+curr_z+curr_fortress);
	*/
	print_perm(source, size);

	found = 0;
	rev_dist++;
	for(i=0;i<size && found == 0;i++)
	for(j=i;j<size && found == 0;j++)
	{
		copy(curr_perm, temp_perm, size);
		reverse(temp_perm, size, i, j);
		next_b = num_breakpoints(temp_perm, size);
		next_c = num_cycles(temp_perm, size); 
		next_z_fortress = is_bad_config(temp_perm, size);
		next_z = next_z_fortress/1000;
		next_fortress = next_z_fortress - 1000*next_z;
		if((curr_b+curr_z-curr_c+curr_fortress)-(next_b+next_z-next_c+next_fortress) == 1)
		{
			reverse(source, size, i, j);
			/*
			printf("reverse [%d %d]\n", i+1,j+1);
			*/
			copy(temp_perm, curr_perm, size);
			curr_b = next_b ;
			curr_c = next_c ;
			curr_z = next_z ;
			curr_fortress = next_fortress ;

			found = 1;
		}
	}
	
	if(found == 0)
	{
		printf("PROBLEM: ");
		print_perm(curr_perm, size);
		exit(1);
	}
}
print_perm(target, size);
printf("\nreversal distance = %d\n",rev_dist);
}
/*-------------return a random permutation of the given size */

void gen_perm(int *perm, int size)
{
int i, num_swaps, index1, index2, sign;
for(i=0;i<size;i++)
	perm[i] = i+1;
num_swaps = rand() % ceil(size/2);
for(i=0;i<num_swaps;i++)
{
int temp;
	index1 = rand() % size;
	index2 = rand() % size;
	temp = perm[index1];
	perm[index1] = perm[index2];
	perm[index2] = temp;
}
for(i=0;i<size;i++)
{
	sign = rand() % 50;
	perm[i] = perm[i] * ((sign == 0)? -1:1) ;
}
}
/*-----------------------------*/

void two2one(int *source, int *target, int *g_perm, int size)
{
int target_index[MAX_SIZE];
int i;

for(i=0;i<size;i++)
	target_index[abs(target[i])] = (i+1)*((target[i]<0)? -1:1);
for(i=0;i<size;i++)
	g_perm[i] = target_index[abs(source[i])] * ((source[i]<0)? -1:1);
}
/*-----------------------------*/
main(int argc, char *argv[])
{
int curr = 0;
int size, index, start, end, to, new_index;
int j,i,fact_size, dist, all_reached;
int _b, _c; 
//++pac 

//printf("enter size : ");
//fflush(stdout);
scanf("%d", &size);
//++pac 
//printf("enter source permutation: ");
//fflush(stdout);
for(i=0;i<size;i++)
	scanf("%d", &source[i]);
//++pac 
//printf("enter target permutation: ");
//fflush(stdout);
for(i=0;i<size;i++)
	scanf("%d", &target[i]);
two2one(source, target, g_perm, size);
reversal_dist(g_perm, size);
}
